/**
 * <copyright>
 *
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id: SetRequiredBundleVersionRanges.java,v 1.3 2006/06/21 20:47:49 marcelop Exp $
 */
package org.eclipse.emf.build.ant.taskdefs;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.regex.Pattern;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.types.FileSet;

import org.eclipse.emf.build.ant.util.AntUtil;
import org.eclipse.emf.build.ant.util.Util;


/**
 * <p>
 * Sets the lower and upper version of each required bundle listed in 
 * a MANIFEST file.
 * </p>
 * <p>
 * <table order="1" cellpadding="2" cellspacing="0">
 * <tr><td><b>Attribute</b></td><td><b>Description</b></td></tr>
 * <tr>
 *    <td valign="top">verbose</td>
 *    <td>Log the files that are changes, the ranges of each dependency and so.</td>
 * </tr>
 * </table> 
 * </p>
 * <p>
 * <table order="1" cellpadding="2" cellspacing="0">
 * <tr><td><b>Nested Element</b></td><td><b>Description</b></td></tr>
 * <tr>
 *    <td valign="top">ManifestSet</td>
 *    <td>A fileset that should be used to specify the MANIFEST.MF files that are 
 *    modified by this task.</td>
 * </tr>
 * <tr>
 *    <td valign="top">DependencySet</td>
 *    <td>A fileset that should be used to specify MANIFEST.MF files or jar files
 *    to be used to query the current version of the required bundles.</td>
 * </tr>
 * <tr>
 *    <td valign="top">VersionRangeRule</td>
 *    <td>Define the version range of the required bundles.  The rules are evaluated
 *    in the order they are defined in the Ant script.<br />
 *    The attributes are:
 *    <table order="1" cellpadding="2" cellspacing="0">
 *      <tr>
 *          <td valign="top">lowerVersion</td>
 *          <td>The lower version of the range. It can be a specific version like "3.0.0" or
 *          "3.1.2.v123" or a formula using the <tt>current</tt> variable like 
 *          "{current+1}.{current-1}.{current}".  The default is<br />
 *          {current}.{current}.{current}</td>
 *      </tr>
 *      <tr>
 *          <td valign="top">lowerIncluded</td>
 *          <td>Whether the lower version is included in the range or not.  The default is
 *          <tt>true</tt>.</td>
 *      </tr>
 *      <tr>
 *          <td valign="top">upperVersion</td>
 *          <td>The upper version of the range. As the lower version, it can be a specific 
 *          version or a formula using the <tt>current</tt> variable.  The default is<br />
 *          {current+1}.0.0</td>
 *      </tr>
 *      <tr>
 *          <td valign="top">upperIncluded</td>
 *          <td>Whether the upper version is included in the range or not.  The default is
 *          <tt>false</tt>.</td>
 *      </tr>
 *      <tr>
 *          <td valign="top">requiredBundleRegex</td>
 *          <td>A regular expression to restrict which required bundles will have their
 *          version range defined by this VersionRangeRule.  The version range rule applies
 *          to all bundles if this attribute is not specified.</td>
 *      </tr>
 *      <tr>
 *          <td valign="top">overwrite</td>
 *          <td>Whether the task should either keep an existing version range or replace it with
 *          the rule definition.  The default value is <tt>false</tt></td>
 *      </tr>
 *    </table>
 *    </td>
 * </tr>
 * </table> 
 * </p>
 * <p>
 * This is an example of how this task can be used in an Ant script:
 * <pre>
 * &lt;emf.build.SetRequiredBundleVersionRanges verbose=&quot;true&quot;&gt;
 *   &lt;manifestSet dir=&quot;${buildDir}&quot; includes=&quot;org.eclipse.*.cheatsheets/META-INF/MANIFEST.MF&quot;/&gt;
 *   &lt;dependencySet dir=&quot;${buildDir}&quot;&gt;
 *     &lt;include name=&quot;org.eclipse.emf*\META-INF/MANIFEST.MF&quot; /&gt;
 *     &lt;include name=&quot;org.eclipse.xsd*\META-INF/MANIFEST.MF&quot; /&gt;
 *     &lt;include name=&quot;org.eclipse*jar&quot; /&gt;
 *   &lt;/dependencySet&gt;
 * 
 *   &lt;versionRangeRule lowerVersion=&quot;{current*}&quot; requiredBundleRegex=&quot;org\.eclipse\.(emf|xsd)&quot; /&gt;
 *   &lt;versionRangeRule overwrite=&quot;true&quot; /&gt;
 * &lt;/emf.build.SetRequiredBundleVersionRanges&gt;
 * </pre>
 * </p>
 *  
 * @since 2.2.0
 */
public class SetRequiredBundleVersionRanges extends BaseTask
{
  public static class VersionFormula
  {
    public static final String CURRENT_TOKEN = "current";

    protected String version;

    protected String major;
    protected String minor;
    protected String micro;
    protected String qualifier;

    public VersionFormula(String version)
    {
      if (version == null)
      {
        throw new BuildException("Version formula cannot be null.");
      }
      else
      {
        version = version.trim();
        if (version.length() > 0)
        {
          this.version = version;
          initialize();
        }
      }
    }

    protected void initialize()
    {
      String[] components = version.split("\\.");
      switch (components.length)
      {
        case 4:
          qualifier = components[3].trim();
        case 3:
          micro = components[2].trim();
        case 2:
          minor = components[1].trim();
        case 1:
          major = components[0].trim();
          break;
        default:
          throw new BuildException("Invalid version formula '" + version + "'.");
      }

      if (version.indexOf("*}") >= 0)
      {
        if (qualifier != null)
        {
          int index = qualifier.indexOf("*}");
          if (index >= 0)
          {
            qualifier = qualifier.substring(0, index) + qualifier.substring(index + 1);
          }
        }

        String currentVar = "{" + CURRENT_TOKEN + "}";
        if (micro != null)
        {
          int index = micro.indexOf("*}");
          if (index >= 0)
          {
            if (qualifier != null)
            {
              throw new BuildException("Invalid version formula '" + version + "'.");
            }
            else
            {
              qualifier = currentVar;
              micro = micro.substring(0, index) + micro.substring(index + 1);
            }
          }
        }

        if (minor != null)
        {
          int index = minor.indexOf("*}");
          if (index >= 0)
          {
            if (qualifier != null || micro != null)
            {
              throw new BuildException("Invalid version formula '" + version + "'.");
            }
            else
            {
              qualifier = currentVar;
              micro = currentVar;
              minor = minor.substring(0, index) + minor.substring(index + 1);
            }
          }
        }

        if (major != null)
        {
          int index = major.indexOf("*}");
          if (index >= 0)
          {
            if (qualifier != null || micro != null || minor != null)
            {
              throw new BuildException("Invalid version formula '" + version + "'.");
            }
            else
            {
              qualifier = currentVar;
              micro = currentVar;
              minor = currentVar;
              major = major.substring(0, index) + major.substring(index + 1);
            }
          }
        }
      }
    }

    public String getMajor()
    {
      return major;
    }

    public String getMinor()
    {
      return minor;
    }

    public String getMicro()
    {
      return micro;
    }

    public String getQualifier()
    {
      return qualifier;
    }

    public boolean hasVariable()
    {
      return version.indexOf('{') >= 0 && version.indexOf('}') >= 0;
    }

    public String toString()
    {
      return version;
    }

    public String toVersion()
    {
      return hasVariable() ? null : version;
    }

    public String toVersion(String base)
    {
      if (hasVariable())
      {
        VersionFormula baseVersion = new VersionFormula(base);
        if (baseVersion.getMajor() == null) baseVersion.major = "0";
        if (baseVersion.getMinor() == null) baseVersion.minor = "0";
        if (baseVersion.getMicro() == null) baseVersion.micro = "0";
        
        String newQualifier = computeComponent(baseVersion.getQualifier(), getQualifier());

        int newMicro = computeComponent(Integer.parseInt(baseVersion.getMicro()), getMicro());
        if (newQualifier != null && newMicro < 0)
        {
          newMicro = 0;
        }

        int newMinor = computeComponent(Integer.parseInt(baseVersion.getMinor()), getMinor());
        if (newMicro >= 0 && newMinor < 0)
        {
          newMinor = 0;
        }

        int newMajor = computeComponent(Integer.parseInt(baseVersion.getMajor()), getMajor());
        if (newMinor >= 0 && newMajor < 0)
        {
          newMajor = 0;
        }

        StringBuffer version = new StringBuffer();
        if (newMajor >= 0) version.append(".").append(newMajor);
        if (newMinor >= 0) version.append(".").append(newMinor);
        if (newMicro >= 0) version.append(".").append(newMicro);
        if (newQualifier != null) version.append(".").append(newQualifier);
        return version.deleteCharAt(0).toString();
      }
      else
      {
        return toVersion();
      }
    }

    protected String computeComponent(String baseComponent, String formulaComponent)
    {
      if (formulaComponent == null)
      {
        return null;
      }
      else if (formulaComponent.charAt(0) == '{' && formulaComponent.charAt(formulaComponent.length() - 1) == '}')
      {
        String variable = formulaComponent.substring(1, formulaComponent.length() - 1);

        if (CURRENT_TOKEN.equals(variable))
        {
          if (baseComponent != null)
          {
            baseComponent = baseComponent.trim();
            if (baseComponent.length() == 0)
            {
              baseComponent = null;
            }
          }
          return baseComponent;
        }
        else
        {
          throw new BuildException("Cannot resolve the version formula component '" + formulaComponent + "'.");
        }
      }
      return formulaComponent;
    }

    protected int computeComponent(int baseComponent, String formulaComponent)
    {
      if (formulaComponent == null)
      {
        return -1;
      }
      else if (formulaComponent.charAt(0) == '{' && formulaComponent.charAt(formulaComponent.length() - 1) == '}')
      {
        String variable = formulaComponent.substring(1, formulaComponent.length() - 1);

        int delta = 0;
        String operation = variable.replaceFirst("^[^+-]*", "").trim();
        if (operation.length() > 0)
        {
          variable = variable.substring(0, variable.length() - operation.length()).trim();
          try
          {
            delta = Integer.parseInt(operation.replace('+', ' ').trim());
          }
          catch (NumberFormatException e)
          {
          }
        }

        if (CURRENT_TOKEN.equals(variable))
        {
          int component = baseComponent + delta;
          if (component < 0)
            component = 0;
          return component;
        }
      }

      try
      {
        return Integer.parseInt(formulaComponent);
      }
      catch (NumberFormatException e)
      {
        throw new BuildException("Cannot resolve the version formula component '" + formulaComponent + "'.");
      }
    }
  }

  public static class VersionRangeRule
  {
    protected VersionFormula lowerVersion = new VersionFormula(new StringBuffer()
      .append("{").append(VersionFormula.CURRENT_TOKEN).append("}")
      .append(".{").append(VersionFormula.CURRENT_TOKEN).append("}")
      .append(".{").append(VersionFormula.CURRENT_TOKEN).append("}").toString());

    protected VersionFormula upperVersion = new VersionFormula(new StringBuffer()
      .append("{").append(VersionFormula.CURRENT_TOKEN).append("+1}.0.0").toString());

    protected boolean overwrite = false;
    protected boolean lowerIncluded = true;
    protected boolean upperIncluded = false;

    protected Pattern requiredBundleRegex;

    public boolean isLowerIncluded()
    {
      return lowerIncluded;
    }

    public void setLowerIncluded(boolean lowerIncluded)
    {
      this.lowerIncluded = lowerIncluded;
    }

    public boolean isUpperIncluded()
    {
      return upperIncluded;
    }

    public void setUpperIncluded(boolean upperIncluded)
    {
      this.upperIncluded = upperIncluded;
    }

    public VersionFormula getLowerVersion()
    {
      return lowerVersion;
    }

    public void setLowerVersion(String version)
    {
      this.lowerVersion = new VersionFormula(version);
    }

    public VersionFormula getUpperVersion()
    {
      return upperVersion;
    }

    public void setUpperVersion(String version)
    {
      this.upperVersion = new VersionFormula(version);
    }

    public boolean isOverwrite()
    {
      return overwrite;
    }

    public void setOverwrite(boolean overwrite)
    {
      this.overwrite = overwrite;
    }

    public void setRequiredBundleRegex(String regex)
    {
      requiredBundleRegex = Pattern.compile(regex);
    }

    public Pattern getRequiredBundleRegex()
    {
      return requiredBundleRegex;
    }

    public String toVersionRange()
    {
      return toString(getLowerVersion().toVersion(), getUpperVersion().toVersion());
    }

    public String toVersionRange(String base)
    {
      return toString(getLowerVersion().toVersion(base), getUpperVersion().toVersion(base));
    }

    public String toString()
    {
      return toString(getLowerVersion().toString(), getUpperVersion().toString());
    }

    protected String toString(String lowerVersion, String upperVersion)
    {
      StringBuffer sb = new StringBuffer();
      sb.append(isLowerIncluded() ? "[" : "(");
      sb.append(lowerVersion);
      sb.append(",").append(upperVersion);
      sb.append(isUpperIncluded() ? "]" : ")");
      return sb.toString();
    }
  }

  protected static final Pattern VERSION_RANGE_PATTERN = Pattern.compile(";\\s*bundle-version\\s*=\\s*\"(.*?)\"", Pattern.MULTILINE);

  protected List versionRangeRules;
  protected boolean hasCatchAllRule = false;
  protected List manifestSets;
  protected List dependencySets;

  protected List manifestFiles;
  protected List unmodifiableManifestFiles;
  protected List jarFiles;

  protected Map symbolicNameCurrentVersionMap;
  protected boolean hasAnalyzedFileLists = false;
  protected Map symbolicNameVersionRangeRuleMap;

  public void addConfiguredVersionRangeRule(VersionRangeRule versionRangeRule)
  {
    assertTrue("You need to specify the upper version of each versionRangeRule", versionRangeRule.getUpperVersion() != null);
    assertTrue("You need to specify the lower version of each versionRangeRule", versionRangeRule.getLowerVersion() != null);
    if (hasCatchAllRule)
    {
      throw new BuildException("Any rule declared after a rule without a required bundle regex attribute will never be used.");
    }
    hasCatchAllRule |= versionRangeRule.getRequiredBundleRegex() == null;    
    
    if (versionRangeRules == null)
    {
      versionRangeRules = new ArrayList();
    }
    versionRangeRules.add(versionRangeRule);
  }

  public void addManifestSet(FileSet manifestSet)
  {
    if (manifestSets == null)
    {
      manifestSets = new ArrayList();
    }
    manifestSets.add(manifestSet);
  }
  
  public void addDependencySet(FileSet dependencySet)
  {
    if (dependencySets == null)
    {
      dependencySets = new ArrayList();
    }
    dependencySets.add(dependencySet);
  }  
  
  protected void checkAttributes() throws BuildException
  {
    assertTrue("You need to specify at least one manifestSet.", manifestSets != null);
    assertTrue("You need to specify at least one versionRangeRule.", versionRangeRules != null);
  }

  protected void doExecute() throws Exception
  {
    computeFiles();

    int size = manifestFiles.size();
    if (size > 0)
    {
      log("Manifest files to change: " + size);

      for (Iterator i = manifestFiles.iterator(); i.hasNext();)
      {
        File manifestFile = (File)i.next();
        log("Analyzing '" + manifestFile.getAbsolutePath() + "'");

        String content = Util.readFile(manifestFile);
        Manifest manifest = new Manifest(new ByteArrayInputStream(content.getBytes()));
        String requiredBundles = manifest.getMainAttributes().getValue("Require-Bundle");
        if (requiredBundles != null)
        {
          String originalContent = content;
          String[] requiredBundleArray = splitRequiredBundles(requiredBundles);
          for (int j = 0; j < requiredBundleArray.length; j++)
          {
            String symbolicName = requiredBundleArray[j].replaceAll("[\\[\\(].*[\\]\\)]", "").trim();
            VersionRangeRule versionRangeRule = selectVersionRangeRule(symbolicName);
            if (!symbolicName.equals(requiredBundleArray[j]) && !versionRangeRule.isOverwrite())
            {
              log("Skipping '" + requiredBundleArray[j] + "'");
              continue;
            }

            String currentVersion = getCurrentVersion(symbolicName);
            String versionRange = versionRangeRule.toVersionRange(currentVersion);
            log("Range of '" + symbolicName + "' (current version '" + currentVersion + "') is '" + versionRange + "'.");

            content = setRequiredBundleVersionRange(content, symbolicName, versionRange);
          }

          if (originalContent.equals(content))
          {
            log("File '" + manifestFile.toString() + "' was not changed by this task.");
          }
          else
          {
            log("Saving '" + manifestFile.toString() + "'");
            Util.writeFile(manifestFile, content);
          }
        }
      }
    }
  }

  protected void computeFiles()
  {
    List manifestFiles = new ArrayList();
    for (Iterator i = manifestSets.iterator(); i.hasNext();)
    {
      FileSet manifestSet = (FileSet)i.next();
      List files = AntUtil.computeExistingFiles(getProject(), manifestSet);
      for (Iterator j = files.iterator(); j.hasNext();)
      {
        File file = (File)j.next();
        if ("MANIFEST.MF".equals(file.getName()))
        {
          manifestFiles.add(file);
        }
      }
    }
    if (!manifestFiles.isEmpty())
    {
      this.manifestFiles = manifestFiles;
    }

    if (dependencySets != null)
    {
      List unmodifiableManifestFiles = new ArrayList();
      List jarFiles = new ArrayList();
      for (Iterator i = dependencySets.iterator(); i.hasNext();)
      {
        FileSet dependencySet = (FileSet)i.next();
        List files = AntUtil.computeExistingFiles(getProject(), dependencySet);
        for (Iterator j = files.iterator(); j.hasNext();)
        {
          File file = (File)j.next();
          String fileName = file.getName();
          if ("MANIFEST.MF".equals(fileName))
          {
            unmodifiableManifestFiles.add(file);
          }
          else if (fileName.endsWith("jar"))
          {
            jarFiles.add(file);
          }
        }
      }
      if (!unmodifiableManifestFiles.isEmpty())
      {
        this.unmodifiableManifestFiles = unmodifiableManifestFiles;
      }
      if (!jarFiles.isEmpty())
      {
        this.jarFiles = jarFiles;
      }
    
      if (manifestFiles != null && unmodifiableManifestFiles != null)
      {
        for (Iterator i = manifestFiles.iterator(); i.hasNext();)
        {
          File manifestFile = (File)i.next();
          if (unmodifiableManifestFiles.remove(manifestFile))
          {
            log("The file '" + manifestFile.getAbsolutePath() + "' will be modified even though it was also selected from an unmodifiable manifestSet.");
          }
        }
      }
    }
  }

  protected String getCurrentVersion(String symbolicName) throws IOException, BuildException
  {
    if (symbolicNameCurrentVersionMap == null)
    {
      symbolicNameCurrentVersionMap = new HashMap();

      if (jarFiles != null)
      {
        for (Iterator i = jarFiles.iterator(); i.hasNext();)
        {
          File file = (File)i.next();
          JarFile jarFile = new JarFile(file);
          Manifest manifest = jarFile.getManifest();
          String[] symbolicNameVersion = computeSymbolicNameVersion(manifest);
          if (symbolicNameVersion != null)
          {
            symbolicNameCurrentVersionMap.put(symbolicNameVersion[0], symbolicNameVersion[1]);
          }
          else
          {
            throw new BuildException("Unable to read the version from '" + file.getAbsolutePath() + "'.");
          }
        }
      }
      
      if (unmodifiableManifestFiles != null)
      {
        symbolicNameCurrentVersionMap.putAll(computeSymbolicNameVersionMap(unmodifiableManifestFiles));
      }
      
      if (manifestFiles != null)
      {
        symbolicNameCurrentVersionMap.putAll(computeSymbolicNameVersionMap(manifestFiles));
      }
    }

    String version = (String)symbolicNameCurrentVersionMap.get(symbolicName);
    if (version != null)
    {
      return version;
    }
    else
    {
      throw new BuildException("Unable to find the version of '" + symbolicName + "'.");
    }
  }

  protected Map computeSymbolicNameVersionMap(Collection manifestFiles) throws IOException, BuildException
  {
    Map symbolicNameCurrentVersionMap = new HashMap();
    for (Iterator i = manifestFiles.iterator(); i.hasNext();)
    {
      File manifestFile = (File)i.next();
      Manifest manifest = new Manifest(new BufferedInputStream(new FileInputStream(manifestFile)));
      String[] symbolicNameVersion = computeSymbolicNameVersion(manifest);
      if (symbolicNameVersion != null)
      {
        symbolicNameCurrentVersionMap.put(symbolicNameVersion[0], symbolicNameVersion[1]);
      }
      else
      {
        throw new BuildException("Unable to read the version from '" + manifestFile.getAbsolutePath() + "'.");
      }
    }
    return symbolicNameCurrentVersionMap;
  }
  
  protected String[] computeSymbolicNameVersion(Manifest manifest)
  {
    String symbolicName = manifest.getMainAttributes().getValue("Bundle-SymbolicName");
    if (symbolicName != null)
    {
      symbolicName = symbolicName.replaceAll(";.*", "").trim();
      String version = manifest.getMainAttributes().getValue("Bundle-Version");
      if (version != null)
      {
        return new String[]{symbolicName, version};
      }
    }
    return null;
  }

  protected String[] splitRequiredBundles(String requiredBundles)
  {
    requiredBundles = VERSION_RANGE_PATTERN.matcher(requiredBundles).replaceAll("$1");
    String[] split = requiredBundles.split(",");
    List result = new ArrayList(split.length);
    for (int i = 0, maxi = split.length; i < maxi; i++)
    {
      String requiredBundle = split[i];
      if (requiredBundle.indexOf('[') >= 0 || requiredBundle.indexOf('(') >= 0)
      {
        if (i + 1 < maxi)
        {
          String aux = split[i + 1];
          if (aux.indexOf(']') >= 0 || aux.indexOf(')') >= 0)
          {
            requiredBundle = requiredBundle + "," + aux;
            i++;
          }
        }
      }
      result.add(requiredBundle.replaceAll(";.*", "").replaceAll("\\s*", ""));
    }
    return (String[])result.toArray(new String [result.size()]);
  }

  protected VersionRangeRule selectVersionRangeRule(String symbolicName)
  {
    if (symbolicNameVersionRangeRuleMap == null)
    {
      symbolicNameVersionRangeRuleMap = new HashMap();
    }
    else
    {
      VersionRangeRule versionRangeRule = (VersionRangeRule)symbolicNameVersionRangeRuleMap.get(symbolicName);
      if (versionRangeRule != null)
      {
        return versionRangeRule;
      }
    }

    for (Iterator i = versionRangeRules.iterator(); i.hasNext();)
    {
      VersionRangeRule versionRangeRule = (VersionRangeRule)i.next();
      if (versionRangeRule.getRequiredBundleRegex() == null || versionRangeRule.getRequiredBundleRegex().matcher(symbolicName).find())
      {
        symbolicNameVersionRangeRuleMap.put(symbolicName, versionRangeRule);
        return versionRangeRule;
      }
    }

    throw new BuildException("Unable to find a versionRangeRule for '" + symbolicName + "'");
  }

  protected String setRequiredBundleVersionRange(String manifestContent, String symbolicName, String versionRange)
  {
    StringBuffer content = new StringBuffer(manifestContent);

    int headerIndex = content.indexOf("Require-Bundle:");
    while (true)
    {
      if (headerIndex < 0)
      {
        throw new BuildException("The manifest must contain a 'RequireBundle' header.");
      }
      else if (headerIndex == 0 || Character.isWhitespace(content.charAt(headerIndex - 1)))
      {
        break;
      }
      else
      {
        headerIndex = content.indexOf("Require-Bundle:", headerIndex);
      }
    }

    int bundleIndex = content.indexOf(symbolicName, headerIndex);
    while (true)
    {
      if (bundleIndex < 0)
      {
        throw new BuildException("The manifest must have '" + symbolicName + "' as a required bundle.");
      }
      
      int afterSymbolicNameIndex = bundleIndex + symbolicName.length();
      if (Character.isWhitespace(content.charAt(bundleIndex-1)))
      {
        if (afterSymbolicNameIndex < content.length())
        {
          char ch = content.charAt(afterSymbolicNameIndex);
          if (Character.isWhitespace(ch) || ch == ';' || ch == ',')
          {
            break;
          }
        }
        else
        {
          break;
        }
      }
      
      bundleIndex = content.indexOf(symbolicName, afterSymbolicNameIndex);
    }

    int endOfBundleIndex = -1;
    boolean isBetweenQuotes = false;
    for (int i = bundleIndex + symbolicName.length(), maxi = content.length(); i < maxi; i++)
    {
      char ch = content.charAt(i);
      switch (ch)
      {
        case '\"':
        {
          isBetweenQuotes = !isBetweenQuotes;
          break;
        }
        
        case ',':
        {
          if (!isBetweenQuotes)
          {
            endOfBundleIndex = i;
            i = maxi;
          }
          break;
        }
        
        case '\r':
        case '\n':
        {
          if (i+1 < maxi && Character.isLetter(content.charAt(i+1)))
          {
            endOfBundleIndex = i;
            i = maxi;            
          }
          break;
        }
      }
    }
    if (endOfBundleIndex < 0)
    {
      endOfBundleIndex = content.length()-1;
    }

    String bundleText = content.substring(bundleIndex, endOfBundleIndex);
    bundleText = VERSION_RANGE_PATTERN.matcher(bundleText).replaceFirst("");
    bundleText = bundleText + ";bundle-version=\"" + versionRange + "\"";

    content.replace(bundleIndex, endOfBundleIndex, bundleText);
    return content.toString();
  }
}
